/* $Id$

   Open - record and display opening times
   Copyright (C) 2001 Fraser McCrossan http://www.joat.ca

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
*/

#include <PalmOS.h>

#include "resource.h" /* resource IDs first */
#include "open.h"

extern Int16 dayofweek;
extern opentimetype minuteofday;

DmOpenRef db; /* the open database */
LocalID dbid; /* ID of the open database */

/* declare these so they can call each other - not needed outside this
   module */
Int16 CompareRemaining(void *, void *);
Int16 CompareAlpha(void *, void *);

DmOpenRef OpenDB(LocalID *rdbid)
{
  Err e;
  LocalID dbid;
  Boolean created;
  DmOpenRef d;
  UInt32 creator;
  UInt16 v = version;
  UInt16 attribs; /* used to fetch and then force set of backup bit */
  UInt16 i;
  MemHandle h;
  openclosev1 *oldp;
  void *newp;
  openclose new;
  Char newname[MAXNAMELENGTH];

  e = DmCreateDatabase(CardNum, DBName, appID, DBType, false /* not a resource DB */);
  if (e == errNone || e == dmErrAlreadyExists) {
    created = (e == errNone); /* need to know this a little lower down */
    /* either of these possibilities are fine, find it's localID */
    if ((dbid = DmFindDatabase(0, DBName)) == 0) {
      return 0; /* can't find it, so can't open it */
    }
    /* set the DB version if we just created it */
    if (e == errNone) {
      (void) DmSetDatabaseInfo(0, dbid, NULL, NULL, &v, NULL,
			       NULL, NULL, NULL, NULL, NULL, NULL, NULL);
    }
    /* since we screwed up in a previous version, check for the old
       creator ID and change to the new one, if necessary; also check
       for a v1 database and upgrade to v2 */
    DmDatabaseInfo(0, dbid, NULL, &attribs, &v, NULL, NULL, NULL,
		   NULL, NULL, NULL, NULL, &creator);
    if (creator != appID) {
      /* we need to change to the official one */
      creator = appID;
    }
    attribs |= dmHdrAttrBackup; /* make sure the database gets backed up */
    DmSetDatabaseInfo(0, dbid, NULL, &attribs, NULL, NULL, NULL, NULL, NULL,
		      NULL, NULL, NULL, &creator);
    /* finally, we get to open it! */
    d = DmOpenDatabase(0, dbid, dmModeReadWrite);
    if (d != 0 && created) {
      LocalID ai;
      MemHandle h;
      /* initialize the AppInfo block with categories */
      if ((ai = DmGetAppInfoID(d)) == 0) {
	if ((h = DmNewHandle(d, sizeof(AppInfoType))) != 0) {
	  ai = MemHandleToLocalID(h);
	  DmSetDatabaseInfo(CardNum, dbid,
			    0, 0, 0, 0,
			    0, 0, 0,
			    &ai, 0, 0, 0);
	}
	else {
	  FrmAlert(AppInfoError);
	}
      }
      if (ai) {
	AppInfoPtr ap = MemLocalIDToLockedPtr(ai, CardNum);
	DmSet(ap, 0, sizeof(AppInfoType), 0);
	CategoryInitialize(ap, CategoryRsrc);
	MemPtrUnlock(ap);
      }
    }
    switch (v) {
    case v1:
      /* upgrade it */
      for (i = 0; i < DmNumRecords(d); ++i) {
	UInt32 newsize;
	UInt16 j;

	h = DmQueryRecord(d, i);
	newsize = MemHandleSize(h) +
	  (sizeof(openclose) - sizeof(openclosev1));
	/* lock the chunks and copy old data to new data, and null out new
	   data not present in old record */
	oldp = MemHandleLock(h);
	for (j = 0; j < DAYSINWEEK; ++j) {
	  new.openat[j] = oldp->openat[j];
	  new.closeat[j] = oldp->closeat[j];
	  new.exceptfrom[j] = new.exceptto[j] = 0; /* null by default */
	}
	new.pastmid = oldp->pastmid;
	StrCopy(newname, (Char *)oldp + sizeof(openclosev1));
	MemHandleUnlock(h);
	h = DmResizeRecord(d, i, newsize);
	newp = MemHandleLock(h);
	DmWrite(newp, 0, &new, sizeof(new));
	DmWrite(newp, sizeof(new), newname, StrLen(newname)+1);
	MemHandleUnlock(h);
	DmReleaseRecord(d, i, true);
      }
      v = version;
      DmSetDatabaseInfo(0, dbid, NULL, NULL, &v, NULL, NULL, NULL, NULL,
			NULL, NULL, NULL, NULL);
      break;
    case version:
      /* do nothing, all is fine */
      break;
    default:
      FrmAlert(DBFailUpgrade);
      DmCloseDatabase(d);
      return 0;
    }
    if (rdbid != NULL) {
      *rdbid = dbid;
    }
    return d;
  }
  else {
    FrmCustomAlert(InternalError, "wierd create error", __FUNCTION__, "");
  }
  return 0; /* can't open it, sorry */
}

void CloseDB(DmOpenRef *dbp)
{
  /* don't check the error code - nothing we can do about it if it fails
     anyway */
  (void) DmCloseDatabase(*dbp);
}

/* compare by name */
Int16 CompareRemaining(void *r1, void *r2)
{
  openclose *o1 = (openclose *)r1;
  openclose *o2 = (openclose *)r2;
  Boolean dummy;
  opentimetype wc1, wc2;

  CheckIfOpen(o1, &dummy, &wc1);
  CheckIfOpen(o2, &dummy, &wc2);
  return wc1 - wc2;
}

Int16 CompareAlpha(void *r1, void *r2)
{
  Char *c1 = (Char *)(r1+sizeof(openclose));
  Char *c2 = (Char *)(r2+sizeof(openclose));

  return StrCaselessCompare(c1, c2);
}

/* sorting comparison routine */
Int16 Compare(void *r1, void *r2, Int16 other,
	      SortRecordInfoPtr si1, SortRecordInfoPtr si2,
	      MemHandle appInfoH)
{
  Int16 rv;
  if (prefs.sortdir == MainSortRemaining) {
    if ((rv = CompareRemaining(r1, r2)) == 0)
      return CompareAlpha(r1, r2);
    return rv;
  }
  else {
    if ((rv = CompareAlpha(r1, r2)) == 0)
      return CompareRemaining(r1, r2);
    return rv;
  }
}

void SortRecords(DmOpenRef *dbp)
{
  DmQuickSort(*dbp, Compare, 0);
}

/* create a new record with the passed information */
Err InsertRecord(DmOpenRef *dbp, openclose *o, Char *n, UInt16 category)
{
  UInt32 size;
  MemHandle new;
  void *tmp;
  UInt16 from = dmMaxRecordIndex;
  UInt16 to;
  UInt16 attr;
  SortRecordInfoType dummy;
  Err e;

  size = sizeof(*o) + StrLen(n) + 1;
  if ((new = DmNewRecord(*dbp, &from, size)) == 0) {
    return DmGetLastErr();
  }
  tmp = MemHandleLock(new);
  if ((e = DmWrite(tmp, 0, o, sizeof(*o))) == errNone) {
    e = DmWrite(tmp, sizeof(*o), n, StrLen(n)+1);
  }
  to = DmFindSortPosition (*dbp, tmp, &dummy, Compare, 0);
  MemHandleUnlock(new);
  DmRecordInfo(*dbp, DmNumRecords(*dbp)-1, &attr, NULL, NULL);
  attr = (attr&dmAllRecAttrs)|(category&dmRecAttrCategoryMask);
  DmSetRecordInfo(*dbp, DmNumRecords(*dbp)-1, &attr, NULL);
  DmReleaseRecord(*dbp, DmNumRecords(*dbp)-1, true);
  DmMoveRecord(*dbp, from, to);
  if (to < DmNumRecords(*dbp)) {
    /* ensure the new record is visible */
    prefs.firstidx = DmPositionInCategory(*dbp, to, prefs.category);
  }
  else {
    prefs.firstidx = to;
  }
  return e;
}

/* update a new record with the passed information */
Err UpdateRecord(DmOpenRef *dbp, UInt16 index, openclose *o, Char *n, UInt16 category)
{
  UInt32 size;
  MemHandle rech;
  void *tmp;
  UInt16 attr;
  Err e;

  size = sizeof(*o) + StrLen(n) + 1;
  /* we ignore the return code here, since we resize it further down anyway */
  if ((rech = DmGetRecord(*dbp, index)) == NULL) {
    return DmGetLastErr();
  }
  if ((rech = DmResizeRecord(*dbp, index, size)) == NULL) {
    return DmGetLastErr();
  }
  tmp = MemHandleLock(rech);
  if ((e = DmWrite(tmp, 0, o, sizeof(*o))) == errNone) {
    e = DmWrite(tmp, sizeof(*o), n, StrLen(n)+1);
  }
  MemHandleUnlock(rech);
  DmRecordInfo(*dbp, index, &attr, NULL, NULL);
  attr = (attr&dmAllRecAttrs)|(category&dmRecAttrCategoryMask);
  DmSetRecordInfo(*dbp, index, &attr, NULL);
  DmReleaseRecord(*dbp, index, true);
  return e;
}

/* translate an on-screen, by category row into a record index */
Err FindIndexFromRow(DmOpenRef *dbp, Int16 row, UInt16 *index)
{
  UInt16 itmp = 0;

  DmSeekRecordInCategory(*dbp, &itmp, *index, dmSeekForward, prefs.category);
  *index = itmp;
  return DmSeekRecordInCategory(*dbp, index, row, dmSeekForward, prefs.category);
}

/* locate the record at the current screen row, lock it and split it
   into components; after using the pointers, the handle must be
   unlocked */
MemHandle QueryRecord(DmOpenRef *dbp, UInt16 index, openclose *o, Char *n)
{
  void *recptr;
  MemHandle rec;

  if ((rec = DmQueryRecord(*dbp, index)) != NULL) {
    recptr = MemHandleLock(rec);
    /* now we have a generic pointer to the whole record, split that
       into pointers to the two sections, and copy to the passed pointers */
    if (o != NULL) {
      MemMove(o, recptr, sizeof(*o));
    }
    if (n != NULL) {
      StrCopy(n, (Char *) (recptr + sizeof(*o)));
    }
    MemHandleUnlock(rec);
    return rec;
  }
  else {
    return 0; /* error if we can't fetch the record */
  }
}

void CheckIfOpen(openclose *oc, Boolean *isopen, opentimetype *willchange)
{
  opentimetype openat, closeat, precloseat; /* these are set depending
					       on the past midnight
					       bit */

  openat = (oc->pastmid & (1<<dayofweek)) ? oc->closeat[dayofweek] : oc->openat[dayofweek];
  closeat = (oc->pastmid & (1<<dayofweek)) ? (oc->openat[dayofweek]+MINUTESINDAY) : oc->closeat[dayofweek];
  /* if the previous day's time crosses midnight, and is *greater than*
      the *current time*, it becomes today's close time */
  if ((precloseat = (oc->pastmid & (1<<((dayofweek-1)%DAYSINWEEK))) ? oc->openat[(dayofweek-1)%DAYSINWEEK] : 0)) {
    if (precloseat > minuteofday) {
      openat = 0;
      closeat = precloseat;
    }
  }
  /* three situations:
     current time in day is less than opening time;
     between opening and closing time;
     after closing time;
     in the third case, we have to search through the week to find
     the next opening time to calculate willchange, otherwise it's easy */
  if (minuteofday < openat) {
    /* we signal that something doesn't open on a particular day by
       having opening and closing times equal to zero, so this
       condition would fail */
    *isopen = false;
    *willchange = openat - minuteofday;
  }
  else if (minuteofday < closeat) {
    *isopen = true;
    *willchange = closeat - minuteofday;
  }
  else {
    UInt16 i;
    *isopen = false;
    /* add the remainder of the day first */
    *willchange = MINUTESINDAY - minuteofday;
    for (i = (dayofweek+1) % DAYSINWEEK;
	 i != dayofweek;
	 i = (i+1) % DAYSINWEEK) {
      openat = (oc->pastmid & (1<<i)) ? oc->closeat[i] : oc->openat[i];
      closeat = (oc->pastmid & (1<<i)) ? (oc->openat[i]+MINUTESINDAY) : oc->closeat[i];
      if (openat == closeat) {
	/* doesn't open on this day; add a day's worth of time */
	*willchange += MINUTESINDAY;
      }
      else {
	/* it does open, add the time of day, then break */
	*willchange += openat;
	break;
      }
    }
    if (i == dayofweek) {
      /* we went right around the loop and found nothing - check if
	 it only opens once a week - that is, if the current time is
	 greater than the opening time - it must also be later than
	 the closing time, else that would have been detected way
	 the hell up there */
      opentimetype openat; /* these are set depending on the past
			      midnight bit */
      openat = (oc->pastmid & (1<<dayofweek)) ? oc->closeat[dayofweek] : oc->openat[dayofweek];
      if (minuteofday > openat) {
	*willchange += openat;
      }
      else {
	*willchange = NEVERCHANGE;
      }
    }
  }
  if (*isopen && oc->exceptfrom[dayofweek] != oc->exceptto[dayofweek]) {
    /* check for the exceptions time, since we are open and it exists */
    if (minuteofday < oc->exceptfrom[dayofweek]) {
      /* before the exception time, we'll change then */
      *willchange = oc->exceptfrom[dayofweek] - minuteofday;
    }
    else if (minuteofday < oc->exceptto[dayofweek]) {
      /* after the start but before the end of the exception time */
      *isopen = false;
      *willchange = oc->exceptto[dayofweek] - minuteofday;
    }
    /* remaining condition has no effect, so ignored */
  }
}
