
#include <PalmOS.h>

#include "resource.h" /* resource IDs first */
#include "open.h"

UInt32 rom_version;

Boolean on_poser;

preftype prefs; /* my preferences! */

Boolean MiscHandler(EventType *e)
{
  switch (e->eType) {

  case frmOpenEvent:
    return false;

  default:
    return false;
  }
}

Boolean AppHandleEvent(EventType *e)
{
  if (e->eType == frmLoadEvent) {
    FormType *form = FrmInitForm(e->data.frmLoad.formID);
    FrmSetActiveForm(form);
    switch (e->data.frmLoad.formID) {

    case MainForm:
      FrmSetEventHandler(form, MainFormHandler);
      return true;

    case AboutForm:
      FrmSetEventHandler(form, AboutFormHandler);
      return true;

    case PlaceForm:
      FrmSetEventHandler(form, PlaceFormHandler);
      return true;

    default:
      FrmCustomAlert(InternalError, "unhandled form", __FUNCTION__, "");
      return false;
    }
    return false;
  }
  else {
    return false;
  }
}

Boolean StartApplication()
{
  UInt16 prefssize;
  //UInt32 secs;
  UInt32 value;
  /* maybe no globals with a Find, so we may need our own variables */
  if (FtrGet(sysFtrCreator, sysFtrNumROMVersion, &rom_version) != 0)
    rom_version = 0;  /* Set to very low version if FtrGet failed.  */
  on_poser = (FtrGet('pose', 0, &value) == 0);
  prefs.category = 0;
  prefs.firstidx = 0;
  prefs.sortdir = MainSortName;
  prefs.hour24 = true;
  prefssize = sizeof(prefs);
  if (PrefGetAppPreferences(appID, 0, &prefs, &prefssize, true)
      == noPreferenceFound) {
    /* if that call failed, nothing we can do about it, so let's assume
       that it's because this is the first launch */
  }
  if ((db = OpenDB(&dbid)) == 0) {
    FrmAlert(DBError);
    return false;
  }
  return true;
}

void EventLoop()
{
  EventType e;
  UInt16 err;
  //UInt32 secs;
  Int32 tickspersec = (Int32) SysTicksPerSecond();
  Int32 timeout;
  /* maybe no globals with a Find, so we may need our own variables */
  //timeout = evtWaitForever;
  timeout = tickspersec;
  do {
    EvtGetEvent(&e, timeout);
    if (!MenuHandleEvent(NULL, &e, &err)) {
      if (!AppHandleEvent(&e)) {
	if (!FrmDispatchEvent(&e)) {
	  SysHandleEvent(&e);
	}
      }
    }
    /* calculate the timeout for the next event so that it falls close to
       the next minute boundary */
    //secs = TimGetSeconds() % 60;
    //timeout = (60 - (Int32)secs) * tickspersec;
  } while (e.eType != appStopEvent);
}

void StopApplication()
{
  FrmCloseAllForms();
  CloseDB(&db);
  /* final dying act, save our preferences */
  PrefSetAppPreferences(appID, 0, 0, (void *) &prefs, sizeof(prefs), true);
}

void Search(FindParamsPtr findParams)
{
  Char s[MAXNAMELENGTH+1]; /* string to copy to */
  MemHandle titleh;
  Char *title;
  Boolean done;
  LocalID dbid;
  DmOpenRef db;
  UInt16 outpos;
  UInt16 recordNum;
  RectangleType r;

  if ((dbid = DmFindDatabase(0, DBName)) == 0) {
    findParams->more = false;
    return; /* can't find it, so can't open it */
  }
  if ((titleh = DmGetResource((DmResType)strRsc, FindTitle))) {
    title = MemHandleLock(titleh);
    done = FindDrawHeader(findParams, title);
    MemHandleUnlock(titleh);
    DmReleaseResource(titleh);
  }
  else {
    done = FindDrawHeader(findParams, "Open (res error)");
  }
  if (done) return;
  db = DmOpenDatabase(0, dbid, findParams->dbAccesMode);
  if (!db) {
    findParams->more = false;
    return;
  }
  recordNum = findParams->recordNum;
  while (true) {
    s[0] = '\0';
    if (!QueryRecord(&db, recordNum, NULL, s)) {
      findParams->more = false;
      break;
    }
    if (FindStrInStr(s, findParams->strToFind, &outpos)) {
      done = FindSaveMatch(findParams, recordNum, outpos, 0, 0, 0, dbid);
      if (done) break;
      FindGetLineBounds(findParams, &r);
      WinDrawChars(s, StrLen(s), r.topLeft.x+1, r.topLeft.y);
      findParams->lineNumber++;
    }
    recordNum++;
  }
  DmCloseDatabase(db);
}

UInt32 PilotMain(UInt16 cmd, void *cmdpbp, UInt16 launchFlags)
{
  Boolean launched = launchFlags & sysAppLaunchFlagNewGlobals;
  /* maybe no globals with a Find, so we may need our own variables */

  switch (cmd) {
  case sysAppLaunchCmdNormalLaunch:
    if (StartApplication()) {
      FrmGotoForm(MainForm);
      EventLoop();
      StopApplication();
    }
    break;

  case sysAppLaunchCmdGoTo:
    if (launched) {
      if (StartApplication()) {
	SetRecord(((GoToParamsPtr)cmdpbp)->recordNum);
	FrmGotoForm(PlaceForm);
	EventLoop();
	StopApplication();
      }
    }
    else { /* already launched */
      SetRecord(((GoToParamsPtr)cmdpbp)->recordNum);
      FrmGotoForm(PlaceForm);
    }
    break;

  case sysAppLaunchCmdFind:
    Search ((FindParamsPtr)cmdpbp);
    break;

  case sysAppLaunchCmdSaveData:
    FrmSaveAllForms();
    break;

  }
  return 0;
}
